from asyncio import run
from typing import Union

from heaven.context import Context
from heaven.response import Response
from heaven.router import App
from heaven.request import Request


class Session(object):
    def __init__(self, headers, params, body):
        pass

    def refresh(self):
        pass


class Earth(object):
    def __init__(self, heaven: App):
        self.__app = heaven
        self.__sessions = {
            "*": Session(headers = None, params = None, body = None)
        }  # here we use keys to save headers etc for reuse

    async def receive(self):
        pass

    def ON(self, middleware: str, replacement: str):
        """
        Replace middlewares you don't want to run otherwise your handlers
        will be invoked with actual middleware definitions.

        Alternatively use custom environmentalal variables to point to new environments.
        """
        pass

    async def send(self):
        pass

    def scope():
        return {}

    @property
    def heaven(self) -> App:
        return self.__app

    def GET(self, endpoint: str, session: Union[str, Session] = None) -> tuple[Request, Response, Context]:
        if isinstance(session, str):
            session = self.__sessions.get(session or '*')
        # write code to simulate GET request to endpoint specified by endpoint and
        # return the trinity of req, res, ctx so you can assert different things with them
        async def runner():
            await self.__app(self.scope(), self.receive(), self.send())
        run(runner())
        req = Request(scope = {}, body = None, receive = None, application = self.__app)
        ctx = Context(self.__app)
        res = Response(self.__app, ctx, req)
        return req, res, ctx
    
    def POST(self, endpoint: str, body = None, session: Sessionable) -> Trinity:
        return req, res, ctx




from http import HTTPStatus
from os import path

from functools import singledispatch, update_wrapper
from http import HTTPStatus
from typing import TYPE_CHECKING

from .constants import MESSAGE_NOT_FOUND, STATUS_NOT_FOUND
from .context import Context
from .tutorials import get_guardian_angel_html, ASYNC_RENDER, NO_TEMPLATING, SYNC_RENDER
from .request import Request
if TYPE_CHECKING:
    from router import App  # pragma: no cover


# For compatibility with older versions of python3 using this
# otherwise would use singledispatchmethod available from ^3.8
def MethodDispatch(method):
    decorated = singledispatch(method)
    def decorator(*args, **kwargs):
        return decorated.dispatch(args[1].__class__)(*args, **kwargs)
    decorator.register = decorated.register
    update_wrapper(decorator, method)
    return decorator


@singledispatch
def _body(payload):
    return payload

@_body.register(str)
def _(payload: str):
    return payload.encode()

@_body.register(int)
@_body.register(float)
def _(payload):
    return f'{payload}'.encode()

def _get_guardian_angel(res: 'Response', error: str, snippet: str):
    res.headers = 'Content-Type', 'text/html'
    res.status = HTTPStatus.INTERNAL_SERVER_ERROR
    res.body = get_guardian_angel_html(error, snippet)


class Response():
    def __init__(self, app: 'App', context: 'Context', request: Request):
        self._app = app
        self._ctx = context
        self._req = request
        self._abort = False
        self._body = MESSAGE_NOT_FOUND.encode()
        self._deferred = []
        self._metadata = {}
        self._headers = []
        self._status = STATUS_NOT_FOUND
        self._template = None
        self._mounted_from_application = None

    @MethodDispatch
    def abort(self, payload):
        self._abort = True
        self._body = payload

    @abort.register(str)
    def _(self, payload: str):
        self._abort = True
        self._body = payload.encode()

    @abort.register(int)
    @abort.register(float)
    def _(self, payload):
        self._abort = True
        self._body = f'{payload}'.encode()

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value):
        self._body = _body(value)

    def defer(self, func) -> 'Response':
        self._deferred.append(func)
        return self

    @property
    def deferred(self):
        return len(self._deferred) > 0

    def header(self, key, val) -> 'Response':
        _encode = lambda k: k.encode('utf-8') if isinstance(k, str) else k
        value = _encode(key), _encode(val)
        self._headers.append(value)
        return self

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, value) -> 'Response':
        key, val = value
        return self.header(key, val)

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        if not isinstance(value, dict): raise ValueError
        self._metadata = value

    def cookie(self, name: str, value: str, **kwargs):
        cookie_string = f'{name}={value}'
        for key, val in kwargs.items():
            _key = {
                'expires': 'Expires',
                'secure': 'Secure',
                'httponly': 'HttpOnly',
                'http_only': 'HttpOnly',
                'samesite': 'SameSite',
                'same_site': 'SameSite',
                'domain': 'Domain',
                'path': 'Path',
                'partitioned': 'Partitioned',
                'max_age': 'Max-Age',
                'maxage': 'Max-Age',
            }.get(key.lower(), key)
            if _key == 'Expires':
                try: val = val.strftime('%a, %d %b %Y %H:%M:%S GMT')
                except: raise ValueError(f'Expires must be a datetime object, got {val}')
            if _key in ['Secure', 'HttpOnly', 'Partitioned']:
                if val: cookie_string += f'; {_key}'
                continue
            if _key == 'Max-Age':
                try: val = int(val)
                except: raise ValueError(f'Max-Age must be an integer, got {val}')
            if _key == 'SameSite':
                _val = str(val).capitalize()
                if _val not in ['Strict', 'Lax', 'None']:
                    raise ValueError(f'SameSite must be one of Strict, Lax, None, got {val}')
                val = _val
            cookie_string += f'; {_key}={val}'
        self.headers = 'Set-Cookie', f'{name}={value}'

    async def render(self, name: str, **contexts) -> 'Response':
        """Serve html file walking up parent router/app tree until base parent if necessary"""
        templater = self._app._templater
        self.headers = 'content-type', 'text/html; charset=utf-8'
        # if self._mounted_from_application: templater = self._mounted_from_application._templater or templater
        if not templater:
            return _get_guardian_angel(self, 'You did not enable templating', NO_TEMPLATING)

        if not templater.is_async:
            return _get_guardian_angel(self, 'Trying to use Sync HTML Renderer to render HTML Async', ASYNC_RENDER)

        template = templater.get_template(name)
        self.body = await template.render_async({'ctx': self._ctx, 'res': self, 'req': self._req, **contexts})
        return self

    def renders(self, name: str, **contexts) -> 'Response':
        """Synchronous version of render method above"""
        templater = self._app._templater
        self.headers = 'content-type', 'text/html; charset=utf-8'
        if not templater:
            return _get_guardian_angel(self, 'You did not enable templating', NO_TEMPLATING)

        if templater.is_async:
            return _get_guardian_angel(self, 'Trying to use Async HTML Renderer to render Sync HTML', SYNC_RENDER)
        template = templater.get_template(name)
        self.body = template.render({'ctx': self._ctx, 'res': self, 'req': self._req, **contexts})
        return self

    def redirect(self, location, permanent=False) -> 'Response':
        if permanent: self.status = HTTPStatus.PERMANENT_REDIRECT
        else: self.status = HTTPStatus.TEMPORARY_REDIRECT
        self.headers = 'Location', location
        return self

    @property
    def status(self): # pragma: no cover
        return self._status

    @status.setter
    def status(self, value: int) -> 'Response':
        self._status = value
        return self

    @property
    def template(self):  # pragma: no cover
        return self._template

    @template.setter
    def template(self, path):
        self._template

    def out(self, status: int, body, headers=None) -> 'Response':
        self.status = status
        self.body = body
        if headers: self.headers = headers
        return self

    async def interpolate(self, name: str, **contexts):
        """Serve html file walking up parent router/app tree until base parent if necessary"""
        templater = self._app._templater
        if not templater: raise AttributeError('Can not interpolate without enabling templating on the heaven application')
        if not templater.is_async: raise AttributeError('Async rendering not supported by sync renderer')
        template = templater.get_template(name)
        return await template.render_async({'ctx': self._ctx, **contexts})




from datetime import datetime, date
from typing import Any, TYPE_CHECKING
from uuid import UUID

from heaven.form import Form
from heaven.utils import Lookup

if TYPE_CHECKING:
    from heaven import Router


class Request:
    def __init__(self, scope, body, receive, metadata=None, application=None):
        self._application = application
        self._body = body
        self._cookies = None
        self._form = None
        self._route = None
        self._receive = receive
        self.__qh = {}
        self.__queryhint = ''
        self._scope = scope
        self._subdomain, self._headers = metadata
        self._params = None
        self._queries = None
        self._dirty = False
        self._queried = False
        self._mounted_from_application = None

    def _parse_qs(self):
        qs = self._scope.get("query_string")
        qsd = {}
        if not qs:
            return qsd
        else:
            qs = qs.decode() if isinstance(qs, bytes) else qs

        query_kv_pairs = qs.split("&")
        for kv_pair in query_kv_pairs:
            try: key, value = kv_pair.split("=")
            except: continue
            current_value = qsd.get(key)
            coercion = self.__qh.get(key)
            if coercion:
                try: value = coercion(value)
                except: pass
            if not current_value:
                qsd[key] = value
            else:
                if isinstance(current_value, list):
                    current_value.append(value)
                else:
                    qsd[key] = [current_value, value]
        return qsd

    @property
    def app(self) -> "Router":
        return self._application

    @property
    def body(self):
        return self._body

    @property
    def cookies(self):
        if not self._cookies:
            csd = {}
            cookiestring = self.headers.get("cookie")
            if not cookiestring:
                self._cookies = csd
                return self._cookies
            cookies = cookiestring.split("; ")
            for cookie in cookies:
                try: k, v = cookie.split("=", 1)
                except: pass
                else: csd[k] = v
            self._cookies = csd
        return self._cookies

    @property
    def form(self) -> "Form":
        if not "multipart/form-data" in self.headers.get("content-type"):
            return None  # currently none
        if self._form is None:
            form = Form(self)
            self._form = form
            return form
        return self._form

    @property
    def headers(self):
        if not self._headers:
            self._headers = {}
            for header in self._scope.get("headers"):
                self._headers[header[0]] = header[1]
        return self._headers

    @property
    def host(self):
        return self.headers.get('host')
    
    @property
    def ip(self):
        address, port = self._scope.get("client")
        return Lookup({'address': address, 'port': port})

    @property
    def qh(self) -> str:
        return self.__queryhint

    @qh.setter
    def qh(self, val: str):
        '''Here we process queryhints so heaven can try to coerce query string values'''
        if self.__queryhint: raise ValueError('Querystring metadata already set')
        booleans = {'false': False, 'true': True, '1': True, 0: False}
        def boolean(v: str) -> bool:
            if not isinstance(v, str): return False
            return booleans.get(v.lower())
        kinds = {
            'int': int,
            'str': str,
            'float': float,
            'datetime': datetime.fromisoformat,
            'date': date.fromisoformat,
            'uuid': UUID,
            'bool': boolean
        }
        for pair in val.split('&'):
            try: k, v = pair.split(':')
            except: continue
            kind = kinds.get(v)
            if kind: self.__qh[k] = kind
        self.__queryhint = val

    @property
    def route(self):
        return self._route

    @property
    def scheme(self):
        return self._scope.get("scheme")

    @property
    def server(self) -> str:
        server = self._scope.get('server')
        return f'{server[0]}:{server[1]}'

    @property
    def method(self):
        return self._scope.get("method")

    @property
    def mounted(self):
        return self._mounted_from_application

    @mounted.setter
    def mounted(self, value: 'Router'):
        self._mounted_from_application

    @property
    def params(self) -> dict:
        if not self._dirty:
            if not self._params: self._params = {}
            self._params = {**self._params}
            self._dirty = True
        return self._params or {}

    @params.setter
    def params(self, pair):
        if not self._params:
            self._params = {}
        key, value = pair
        self._params[key] = value

    @property
    def queries(self):
        if not self._queried:
            if not self._queries: self._queries = {}
            self._queries = {**self._parse_qs()}
            self._queried = True
        return self._queries or {}

    @queries.setter
    def queries(self, pair):
        if not self._queries:
            self._queries = {}
        self._queries[pair[0]] = pair[1]

    @property
    def querystring(self):
        return self._scope.get("query_string", "")

    @property
    def scheme(self):
        return self._scope.get("scheme")

    @property
    def subdomain(self):
        return self._subdomain

    @property
    def url(self):
        return self._scope.get("path")

    async def stream(self):
        """TODO: Revisit and implement with appropriate testing"""
        return await self._receive()




from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from heaven.request import Request


ST = b'--'
NL = b'\r\n'


def _reqparser(req) -> dict:
    data = {}
    body = req.body
    ct = req.headers.get('content-type')
    stopper = ct.split('boundary=')[-1]
    stopped = body.split(stopper.encode())
    portions = [p.strip(NL).strip(ST).strip(NL) for p in stopped]
    for portion in portions:
        if portion == b'': continue
        prepart, _, postpart = portion.partition(NL * 2)
        key = _keyxtractor(prepart)
        data[key.decode()] = postpart
    return data


def _keyxtractor(s) -> str:
    for portion in s.split(b';'):
        name = portion.split(b'=')
        if name[0].strip(b' ') == b'name': return name[-1].strip(b'"')


class Form(object):
    def __init__(self, req: 'Request'):
        self._data = _reqparser(req)

    def __getattr__(self, name: str) -> Any:
        return self._data.get(name)
